/*

*/

import tokenizer.*;

/**
   Expression Language E11 (compare with E3).

      E ::= E '+' T    // left recursion means left associative
          | E '-' T    // left recursion means left associative
          | T

      T ::= T '*' F    // left recursion means left associative
          | T '/' F    // left recursion means left associative
          | F

      F ::= '-' F      // right recursion means right associative
          | '+' F      // right recursion means right associative
          | '(' E ')'
          | N

      N ::= N D       // "N" for Number, "D for Digit
          | D

      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   We can rewrite the grammar this way.

      E ::= T ( ( '+' | '-' ) T )*   // sum of Terms (iteration)
      T ::= F ( ( '*' | '/' ) F )*   // product of Factors (iteration)
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | N
      N ::= D ( D )*
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   This parser returns the value of the expression.
   It will throw an error if the string cannot be parsed.

   Some example strings are:
      "-5"
      "+5"
      "+(-5)"
      "-(+5)"
      "+-(-+5)"
      "(+2 + -3)"
      "-(4 + 3) / -2"
      "-2 * -(3 + 4)"
      "(2 + 3) * -(4 / 2)"
      "-10 / (1 + -(2 * -3))"
      "-(-(-2 + -3) * -4) + -5"
   This grammar can handle any valid four function
   arithmetic expression.
*/
public class Parse2Value_E11v2
{
   public static int evalE11(String stringInE11)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE11);
      // Parse the list of tokens.
      final int result = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return result;
   }


   /**
      E ::= T ( ( '+' | '-' ) T )*
   */
   private static int getE(Tokenizer tokens)
   {
      int r = getT(tokens);
      while ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("+")
                                   || tokens.peek().lexeme.equals("-") ) )
      {
         if ( tokens.match("+") )
         {
            r += getT(tokens);
         }
         else if ( tokens.match("-") )
         {
            r -= getT(tokens);
         }
      }
      return r;
   }


   /**
      T ::= F ( ( '*' | '/' ) F )*
   */
   private static int getT(Tokenizer tokens)
   {
      int r = getF(tokens);
      while ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("*")
                                   || tokens.peek().lexeme.equals("/") ) )
      {
         if ( tokens.match("*") )
         {
            r *= getF(tokens);
         }
         else if ( tokens.match("/") )
         {
            r /= getF(tokens);
         }
      }
      return r;
   }


   /**
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | N
   */
   private static int getF(Tokenizer tokens)
   {
      final int r;
      if ( tokens.match("+") )
      {
         r = getF(tokens);
      }
      else if ( tokens.match("-") )
      {
         r = - getF(tokens);
      }
      else if ( tokens.match("(") )
      {
         r = getE(tokens);
         if ( ! tokens.match(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
      }
      else
      {
         r = getN(tokens);
      }
      return r;
   }


   /**
      N ::= D ( D )*
   */
   private static int getN(Tokenizer tokens)
   {
      int result = getD(tokens);
      while ( ! tokens.isEmpty()
           && Character.isDigit(tokens.peek().lexeme.charAt(0)) )
      {
         int d = getD(tokens);
         result = (10 * result) + d;
      }
      return result;
   }


   /**
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
   */
   private static int getD(Tokenizer tokens)
   {
      final Token tk = tokens.next();

      if ( ! Character.isDigit(tk.lexeme.charAt(0)) )
         throw new TokenizeError("expected digit, found " + tk);

      return Integer.parseInt(tk.lexeme);
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "-5",
                              "+5",
                              "+(-5)",
                              "-(+5)",
                              "+-(-+5)",
                              "(+2 + -3)",
                              "-(4 + 3) / -2",
                              "-2 * -(3 + 4)",
                              "(2 + 3) * -(4 / 2)",
                              "-10 / (1 + -(2 * -3))",
                              "-(-(-2 + -3) * -4) + -5",
                              "-2 * -(-3 + -4) / +2",
                              "-10 / -(+(-1 - -2)) * -3",
                              "2 - (3 * 4) + 5",
                              "2 * (3 + 4) - 5",
                              "2 * (3 + 4) - -5",
                              "-(2 + 3 + 4) - 5",
                              "2 + -(3 * 4)) + 5",
                              "2 + -(-(3 * 4) + 5"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            final int r = evalE11( str );
            System.out.println( " ... Parsed" );
            // Print the result.
            System.out.println( r + " = " + str );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
