/*

*/

import tokenizer.*;

/**
   Expression Language E2.

      E ::= N '+' E   // right recursion means right associative
          | N

      N ::= N D       // "N" for Number, "D for Digit
          | D

      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   We can rewrite the grammar this way.

      E ::= N [ '+' E ]   // right recursion means right associative
      N ::= D ( D )*
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   This parser returns the value of the expression.
   It will throw an error if the string cannot be parsed.

   Some example strings are:
      "23"
      "23 + 32"
      "12 + 34 + 56"
      "123 + 456 + 789 + 1000"
*/
public class Parse2Value_E2
{
   public static int evalE2(String stringInE2)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE2);
      // Parse the list of tokens.
      final int result = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return result;
   }


   /**
      E ::= N [ '+' E ]
   */
   private static int getE(Tokenizer tokens)
   {
      int r = getN(tokens);
      if ( ! tokens.isEmpty() && tokens.match("+") )
      {
         r += getE(tokens); // recursion
      }
      return r;
   }


   /**
      N ::= D ( D )*
   */
   private static int getN(Tokenizer tokens)
   {
      int result = getD(tokens);
      while ( ! tokens.isEmpty()
           && Character.isDigit(tokens.peek().lexeme.charAt(0)) )
      {
         int d = getD(tokens);
         result = (10 * result) + d;
      }
      return result;
   }


   /**
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
   */
   private static int getD(Tokenizer tokens)
   {
      final Token tk = tokens.next();

      if ( ! Character.isDigit(tk.lexeme.charAt(0)) )
         throw new TokenizeError("expected digit, found " + tk);

      return Integer.parseInt(tk.lexeme);
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "23",
                              "23 + 32",
                              "12 + 34 + 56",
                              "123 + 456 + 789 + 1000",
                              "1+2+3-4",
                              "1 + + 2 + 3",
                              "1 + 2 3 + 4",
                              "1 + 2 +",
                              "+ 1 + 2"
                             };

      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            final int r = evalE2( str );
            System.out.println( " ... Parsed" );
            // Print the result.
            System.out.println( r + " = " + str );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
