/*

*/

import tokenizer.*;

/**
   Expression Language E3.

      E ::= E '+' N   // left recursion means left associative
          | N

      N ::= N D       // "N" for Number, "D for Digit
          | D

      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   We cannot parse this language as it is written using a
   recursive descent parser. The left recursion in the grammar
   causes infinite recursion in the recursive descent parser.
   We must rewrite the grammar to remove the left recursion
   and replace the recursion with a kind of iteration.

   The above grammar can be rewritten to use the Kleene star
   (from "regular expressions") in place of the left recursion.
   The new grammar needs only one production.

      E ::= N ( '+' N )*   // sum of numbers (iteration)
      N ::= D ( D )*
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   This parser returns the value of the expression.
   It will throw an error if the string cannot be parsed.

   Some example strings are:
      "15"
      "12 + 32"
      "1 + 2 + 3"
      "10 + 20 + 30 + 40"
*/
public class Parse2Value_E3
{
   public static int evalE3(String stringInE3)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE3);
      // Parse the list of tokens.
      final int result = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return result;
   }


   /**
      E ::= N ( '+' N )*
   */
   private static int getE(Tokenizer tokens)
   {
      int r = getN(tokens);
      while ( ( ! tokens.isEmpty()) && tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next(); // consume the "+"
         r += getN(tokens); // recursion
      }
      return r;
   }


   /**
      N ::= D ( D )*
   */
   private static int getN(Tokenizer tokens)
   {
      int result = getD(tokens);
      while ( ! tokens.isEmpty()
           && Character.isDigit(tokens.peek().lexeme.charAt(0)) )
      {
         int d = getD(tokens);
         result = (10 * result) + d;
      }
      return result;
   }


   /**
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
   */
   private static int getD(Tokenizer tokens)
   {
      final Token tk = tokens.next();

      if ( ! Character.isDigit(tk.lexeme.charAt(0)) )
         throw new TokenizeError("expected digit, found " + tk);

      return Integer.parseInt(tk.lexeme);
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "15",
                              "12 + 32",
                              "1 + 2 + 3",
                              "10 + 20 + 30 + 40",
                              "10+20+30+40",
                              "1+2+3-4",
                              "10 + 20 +",
                              "+ 11 + 22"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            final int r = evalE3( str );
            System.out.println( " ... Parsed" );
            // Print the result.
            System.out.println( r + " = " + str );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
