/*

*/

import tokenizer.*;

/**
   Expression Language E8 but changed to be right associative.

      E ::= T '+' E   // right recursion means right associative
          | T

      T ::= N '*' T   // right recursion means right associative
          | N

      N ::= N D       // "N" for Number, "D for Digit
          | D

      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   We can rewrite the grammar this way.

      E ::= T [ '+' E ]   // right recursion means right associative
      T ::= N [ '*' T ]   // right recursion means right associative
      N ::= D ( D )*
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   This parser returns the value of the expression.
   It will throw an error if the string cannot be parsed.

   Some example strings are:
      "10 + 200"
      "3 * 5"
      "2 + 3 * 4"
      "2 * 3 + 4"
      "2 + 3 * 4 + 5"
      "2 * 3 + 4 * 5"
      "2 * 3 * 4 * 5"
      "2 * 3 * 4 + 5 * 6"
*/
public class Parse2Value_E8
{
   public static int evalE8(String stringInE8)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE8);
      // Parse the list of tokens.
      final int result = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return result;
   }


   /**
      E ::= T [ '+' E ]
   */
   private static int getE(Tokenizer tokens)
   {
      int r = getT(tokens);
      if ( ! tokens.isEmpty() && tokens.match("+") )
      {
         r += getE(tokens); // recursion
      }
      return r;
   }


   /**
      T ::= N [ '*' T ]
   */
   private static int getT(Tokenizer tokens)
   {
      int r = getN(tokens);
      if ( ! tokens.isEmpty() && tokens.match("*") )
      {
         r *= getT(tokens); // recursion
      }
      return r;
   }


   /**
      N ::= D ( D )*
   */
   private static int getN(Tokenizer tokens)
   {
      int result = getD(tokens);
      while ( ! tokens.isEmpty()
           && Character.isDigit(tokens.peek().lexeme.charAt(0)) )
      {
         int d = getD(tokens);
         result = (10 * result) + d;
      }
      return result;
   }


   /**
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
   */
   private static int getD(Tokenizer tokens)
   {
      final Token tk = tokens.next();

      if ( ! Character.isDigit(tk.lexeme.charAt(0)) )
         throw new TokenizeError("expected digit, found " + tk);

      return Integer.parseInt(tk.lexeme);
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "10 + 200",
                              "3 * 5",
                              "2 + 3 * 4",
                              "2 * 3 + 4",
                              "2 + 3 * 4 + 5",
                              "2 * 3 + 4 * 5",
                              "2 * 3 * 4 * 5",
                              "2 * 3 * 4 + 5 * 6",
                              "2 * + 3",
                              "+ 2 * 3",
                              "* 2 * 3",
                              "2 + + 3",
                              "2 * * 3",
                              "2 * 3 * ",
                              "2 + 3 * ",
                              "2 * 3 + "
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            final int r = evalE8( str );
            System.out.println( " ... Parsed" );
            // Print the result.
            System.out.println( r + " = " + str );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
