/*

*/

import tokenizer.*;


/**
   Expression Language E9 but changed to be right associative.

      E ::= T '+' E    // right recursion means right associative
          | T

      T ::= F '*' T    // right recursion means right associative
          | F

      F ::= '(' E ')'
          | N

      N ::= N D       // "N" for Number, "D for Digit
          | D

      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   We can rewrite the grammar this way.

      E ::= T [ '+' E ]   // right recursion means right associative
      T ::= F [ '*' T ]   // right recursion means right associative
      F ::= '(' E ')'
          | 'a'
      N ::= D ( D )*
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   This parser returns the value of the expression.
   It will throw an error if the string cannot be parsed.

   Some example strings are:
      "(54)"
      "(54 + 45)"
      "(2 + 3) * 4"
      "2 * (3 + 4)"
      "(2 + 3) * (4 + 5)"
      "2 * (3 + (4 * 5))"
      "((2 + 3) * 4) + 5"
      "2 * (3 + 4) * 5"
*/
public class Parse2Value_E9
{
   public static int evalE9(String stringInE9)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE9);
      // Parse the list of tokens.
      final int result = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return result;
   }


   /**
      E ::= T [ '+' E ]
   */
   private static int getE(Tokenizer tokens)
   {
      int r = getT(tokens);
      if ( ! tokens.isEmpty() && tokens.match("+") )
      {
         r += getE(tokens); // recursion
      }
      return r;
   }


   /**
      T ::= F [ '*' T ]
   */
   private static int getT(Tokenizer tokens)
   {
      int r = getF(tokens);
      if ( ! tokens.isEmpty() && tokens.match("*") )
      {
         r *= getT(tokens); // recursion
      }
      return r;
   }


   /**
      F ::= '(' E ')'
          | 'a'
   */
   private static int getF(Tokenizer tokens)
   {
      final int r;
      if ( tokens.match("(") )
      {
         r = getE(tokens);
         if ( ! tokens.match(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
      }
      else
      {
         r = getN(tokens);
      }
      return r;
   }


   /**
      N ::= D ( D )*
   */
   private static int getN(Tokenizer tokens)
   {
      int result = getD(tokens);
      while ( ! tokens.isEmpty()
           && Character.isDigit(tokens.peek().lexeme.charAt(0)) )
      {
         int d = getD(tokens);
         result = (10 * result) + d;
      }
      return result;
   }


   /**
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
   */
   private static int getD(Tokenizer tokens)
   {
      final Token tk = tokens.next();

      if ( ! Character.isDigit(tk.lexeme.charAt(0)) )
         throw new TokenizeError("expected digit, found " + tk);

      return Integer.parseInt(tk.lexeme);
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "(54)",
                              "(54 + 45)",
                              "(2 + 3) * 4",
                              "2 * (3 + 4)",
                              "(2 + 3) * (4 + 5)",
                              "2 * (3 + (4 * 5))",
                              "((2 + 3) * 4) + 5",
                              "2 * (3 + 4) * 5",
                              "2 * ((3 + 4)) * 5",
                              "2 - (3 * 4) + 5",
                              "2 * (3 + 4) - 5",
                              "2 + (3 * 4)) + 5",
                              "2 + ((3 * 4) + 5",
                              "(2 + 3 + 4) - 5"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            final int r = evalE9( str );
            System.out.println( " ... Parsed" );
            // Print the result.
            System.out.println( r + " = " + str );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
